/*

File:   c_utils.c
Purpose: utility library

Copyright Fabrice Nicol <fabnicol@users.sourceforge.net>, 2008

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#ifdef __WIN32__
#include <io.h>
#endif
#include <dirent.h>
#include <stdarg.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <ctype.h>

/* foutput is a public macro
 * foutput_c is a private one
 * using both public and private macros */
//#include "ports.h"
#include "c_utils.h"
#include "private_c_utils.h"

#undef foutput
#define foutput  foutput_c
#undef globals
#define globals  globals_c


/* global parameters in libc_utils may be independently set par initialise_c_utils */

global_utils globals_c;


void initialise_c_utils(_Bool a, _Bool b, _Bool c,  FILE* f)
{
  globals_c.silence=a;

  globals_c.logfile=b;
  globals_c.debugging=c;
  globals_c.journal=f;
}


void pause_dos_type()
{
  char reply;
  do
    {
      fprintf(stderr, "\n%s\n", "Press Enter to continue...");
      scanf("%c", &reply);
      fflush(stdin);
    }
  while (0);
}


/*********************************************************************************************************
 * function: rm_empty_backup_dir
 *   securely removes empty backup directories, doing security tests.
 **********************************************************************************************************/


int rm_empty_backup_dir(const char * rep)
{
  DIR *dir;
  if (((dir=opendir(rep)) != NULL) &&   (rmdir(rep) == -1))
    return(0);
  else
    return(1);

  closedir(dir) ;
  errno=0 ;
  return(0);
}

/*********************************************************************************************************
 * function: clean_exit
 *   logs time; flushes all streams;  erase empty backup dirs; closes log;
 *   pauses before exiting, with different behaviour on Linux and Windows.
 **********************************************************************************************************/

void clean_exit(int message, const char *default_directory)
{

  fflush(stdout);

  rm_empty_backup_dir(default_directory);

  if (globals_c.logfile)
    {
      fflush(globals_c.journal);
      fclose(globals_c.journal);
    }

  exit(message);
}

/****************************************************************
* function: secure_mkdir
*   Creates directories.
****************************************************************/

int secure_mkdir (const char *path, mode_t mode, const char* default_directory)
{

  int i=0, len;
  len = strlen (path);

  // requires std=c99


  if  ((len<=1) && (globals_c.debugging))
    {
      foutput_c("%s\n","[ERR]  Path length could not be allocated by secure_mkdir:\n       Your compiler may not be C99-compliant");
      foutput_c("       path=%s, length=%d\n",path,len);
      clean_exit(EXIT_FAILURE,default_directory);
    }

  char d[len+1];
  memset(d, 0, len+1);

  memmove(d, path, len+1);

  if (d == NULL)
    {
      perror("[MSG] Error: could not allocate directory path string");
      pause_dos_type();
      clean_exit(EXIT_FAILURE, default_directory);
    }


  for (i = 1; i < len; i++)
    {
      if (('/' == d[i]) || ('\\' == d[i]))
        {
#if defined __WIN32__ || defined __CYGWIN__
          if (d[i-1] == ':') continue;
#endif
          d[i] = '\0';

          if ((MKDIR(d, mode) == -1) && (EEXIST != errno))

            {
              foutput_c( "Impossible to create directory '%s'\n", d);
              foutput_c("%s", "Backup directories will be used.\n " );
              pause_dos_type();
              exit(EXIT_FAILURE);

            }
          d[i] = '/';


        }

    }
  // loop stops before end of string as dirpaths can optionally end in '/' under *nix

  if  (MKDIR(path, mode) == -1)
    {
      if (EEXIST == errno)
        {
          errno=0;
          foutput_c("[MSG]  Output directory '%s' already created\n", path);
          return(0);
        }
      foutput_c("[ERR]  Impossible to create directory '%s'\n", path);
      foutput_c("       permission was: %d\n       %s\n", mode, strerror(errno));
      foutput_c( "%s", "       Backup directories will be used.\n");
      errno=0;
      pause_dos_type();
      exit(EXIT_FAILURE);
    }

  return(errno);
}


void starter(compute_t *timer)
{
// The Mingw compiler does not support getrusage
#ifndef __MINGW32__
  getrusage(RUSAGE_SELF, timer->start);
  getrusage(RUSAGE_SELF, timer->nothing);
  timer->nothing->ru_utime.tv_sec -= timer->start->ru_utime.tv_sec;
  timer->nothing->ru_stime.tv_sec -= timer->start->ru_stime.tv_sec;

#endif

}

void print_commandline(int argc, char * const argv[])
{
  int i=0;

  foutput_c("%s \n", "[INF]  Running:");

  for (i=0; i < argc ; i++)
    foutput_c(" %s  ",  argv[i]);
  foutput_c("%s", "\n\n");

}

char* print_time(int verbose)
{
  char outstr[200];
  time_t t;
  struct tm *tmp;

  t = time(NULL);
  tmp = localtime(&t);
  if (tmp == NULL)
    {
      perror("localtime");
      exit(EXIT_FAILURE);
    }


  if (verbose)
    {
      if (strftime(outstr, sizeof(outstr), "%d %b, %Hh %Mm %Ss", tmp) == 0)
        {
          foutput_c("%s\n", "strftime returned 0");
          exit(EXIT_FAILURE);
        }
    }

  else
    if (strftime(outstr, sizeof(outstr), "%Hh %Mm %Ss", tmp) == 0)
      {
        foutput_c("%s\n", "strftime returned 0");
        exit(EXIT_FAILURE);
      }

  if (verbose)
    {
      foutput_c("\nCurrent time is: %s", outstr);
      return NULL;
    }
  else
    return(strdup(outstr));
}

void change_directory(const char * filename)
{
  if (chdir(filename) == -1)
    {
      if (errno == ENOTDIR)
        foutput_c("[ERR]  %s is not a directory\n", filename);
      foutput_c("[ERR]  input file %s does not comply with chdir  specifications\n.", filename);
      pause_dos_type();
      exit(EXIT_FAILURE);
    }
}

int copy_directory(const char* src, const char* dest, mode_t mode)
{

  DIR *dir_src;

  struct stat buf;
  struct dirent *f;
  char path[BUFSIZ];

  if (stat(dest, &buf) == -1)
    {
      perror("[ERR]  copy_directory could not stat file");
      exit(EXIT_FAILURE);
    }


  foutput_c("%c", '\n');

  if (globals_c.debugging)  foutput_c("%s%s\n", "[INF]  Creating dir=", dest);

  secure_mkdir(dest, mode, "./");

  if (globals_c.debugging)   foutput_c("[INF]  Copying in %s ...\n", src);
  change_directory(src);

  dir_src=opendir(".");

  while ((f=readdir(dir_src)) != NULL)
    {
      if (f->d_name[0] == '.') continue;
      STRING_WRITE(path, "%s%c%s", dest, '/', f->d_name)
      if (stat(f->d_name, &buf) == -1)
        {
          perror("[ERR] stat ");
          exit(EXIT_FAILURE);
        }

      /*  Note: on my implementation of Linux (Ubuntu), S_ISDIR and S_ISREG(buf.st_mode) are seemingly buggy
       *  Resorting to masks S_ISDIR and S_IFREG as a way out */

      if (S_IFDIR & buf.st_mode)
        {

          if (globals_c.debugging) foutput_c("%s %s %s %s\n", "[INF]  Copying dir=", f->d_name, " to=", path);

          errno=copy_directory(f->d_name, path, mode);
          continue;
        }
      if (S_IFREG & buf.st_mode)
        {
          if (globals_c.debugging) foutput_c("%s%s to= %s\n", "[INF]  Copying file=", f->d_name, path);
          errno=copy_file(f->d_name, path);
        }
      /* does not copy other types of files(symlink, sockets etc). */

      else continue;
    }
  change_directory("../");
  if (globals_c.debugging)   foutput_c("%s", "[INF]  Done. Backtracking... \n\n");
  closedir(dir_src);
  return(errno);
}

int copy_file_no_p(FILE *infile, FILE *outfile)
{


  char buf[BUFSIZ];
  clearerr(infile);
  clearerr(outfile);
  size_t chunk;



  while (!feof(infile))
    {

      chunk=fread(buf, sizeof(char), BUFSIZ, infile);

      if (ferror(infile))
        {
          fprintf(stderr, "Read error\n");
          return(-1);
        }


      fwrite(buf, chunk* sizeof(char), 1 , outfile);

      if (ferror(outfile))
        {
          fprintf(stderr, "Write error\n");
          return(-1);
        }
    }
  return(0);
}


// Adapted from Yve Mettier's O'Reilly "C en action" book, chapter 10.

int copy_file(const char *existing_file, const char *new_file)
{

  FILE *fn, *fe;
  int errorlevel;

  if (NULL == (fe = fopen(existing_file, "rb")))
    {
      fprintf(stderr, "[ERR]  Impossible to open file '%s' \n in read mode ", existing_file);
      return(-1);
    }
  if (NULL == (fn = fopen(new_file, "wb")))
    {
      fprintf(stderr, "[ERR]  Impossible to open file '%s' in write mode\n", new_file);
      fclose(fe);
      return(-1);
    }


  errorlevel=copy_file_no_p(fe, fn);
  fclose(fe);
  fclose(fn);

  return(errorlevel);

}

int copy_file_p(FILE *infile, FILE *outfile, uint32_t position,uint64_t output_size)
{


  char buf[BUFSIZ];
  clearerr(infile);
  clearerr(outfile);
  size_t chunk=0;
  uint64_t count=0;

  if (fseek(infile, position, SEEK_SET) == -1) return -1;
  end_seek(outfile);

  while ((output_size > count)&&(!feof(infile)))
    {

      chunk= ((output_size-count) < BUFSIZ) ? fread(buf, sizeof(char), (size_t) output_size -count, infile) : fread(buf, sizeof(char), BUFSIZ, infile) ;

      count+=chunk*sizeof(char);


      if (ferror(infile))
        {
          fprintf(stderr, "Read error\n");
          return(-1);
        }


      fwrite(buf, chunk* sizeof(char), 1 , outfile);


      if (ferror(outfile))
        {
          fprintf(stderr, "Write error\n");
          return(-1);
        }
    }
  return((count < output_size) ? PAD : NO_PAD);
}



int get_endianness()
{
  long i=1;
  const char *p=(const char *) &i;
  if (p[0] == 1) return LITTLE_ENDIAN;
  return BIG_ENDIAN;
}


ALWAYS_INLINE_GCC inline uint8_t read_info_chunk(char* pt, char* chunk)
{
    pt+=4;
    /* there may be non-printable characters after I... info chunk labels */
    while (!isprint(*pt)) pt++;
        
    /* this functin should be called only when it it certain that there will be at least one 0 in the area pointed to by pt within the size of chunk */
    int result=snprintf(chunk, MAX_LIST_SIZE, "%s", pt);
    if (result >0) 
       return(1);
    else 
       return(0);
            
}

void parse_wav_header(FILE * infile, infochunk* ichunk)
{
  uint8_t haystack[MAX_HEADER_SIZE]={0};
  int count;
  if ((count=fread(haystack, 1, MAX_HEADER_SIZE, infile)) != MAX_HEADER_SIZE)
    {
      foutput_c("[ERR]  Could not read %d characters from input file\n", MAX_HEADER_SIZE);
      foutput_c("[ERR]  Just read %d\n", count);
      ichunk->span=0;
      return;
    }
  uint8_t *pt=&haystack[0];
  uint8_t span=0;

  fseek(infile, 0, SEEK_SET);

  do
    {
      if ((pt=memchr(haystack+span+1, 'd', MAX_HEADER_SIZE-1-span)) == NULL)
        {
          foutput_c("[ERR]  Could not find substring 'data' among %d characters\n", MAX_HEADER_SIZE);
          if (globals_c.debugging)
            {
              hexdump_header(infile, MAX_HEADER_SIZE);
            }
          ichunk->span=0;
          return;
        }

      span=pt-haystack;

      if ((*(pt + 1) == 'a') && (*(pt + 2) == 't') && (*(pt + 3) == 'a')) break;

    }
  while ( span < MAX_HEADER_SIZE-7);

pt=&haystack[0];

if (span > 36) 
{
    /* header is non-standard, looking for INFO chunks */
    /* The RIFF standard defines the following chunks, of which only INAM, IART, ICMT, ICOP, ICRD, IGNR will be parsed */
    /*
  IARL  Archival Location. Indicates where the subject of the file is archived.
  IART  Artist. Lists the artist of the original subject of the file. For example, "Michaelangelo."
  ICMS  Commissioned. Lists the name of the person or organization that commissioned the subject of the file. For example, "Pope Julian II."
  ICMT  Comments. Provides general comments about the file or the subject of the file. If the comment is several sentences long, end each sentence with a period. Do not include newline characters.
  ICOP  Copyright. Records the copyright information for the file. For example, "Copyright Encyclopedia International 1991." If there are multiple copyrights, separate them by a semicolon followed by space.
  ICRD  Creation date. Specifies the date the subject of the file was created. List dates in year-month-day format, padding one-digit months and days with a zero on the left. For example, "1553-05-03" for May 3, 1553.
  ICRP  Cropped. Describes whether an image has been cropped and, if so, how it was cropped. For example, "lower right corner."
  IDIM  Dimensions. Specifies the size of the original subject of the file. For example, "8.5 in h, 11 in w."
  IDPI  Dots Per Inch. Stores dots per inch setting of the digitizer used to produce the file, such as "300."
  IENG  Engineer. Stores the name of the engineer who worked on the file. If there are multiple engineers, separate the names by a semicolon and a blank. For example, "Smith, John; Adams, Joe."
  IGNR  Genre. Describes the original work, such as, "landscape," "portrait," "still life," etc.
  IKEY  Keywords. Provides a list of keywords that refer to the file or subject of the file. Separate multiple keywords with a semicolon and a blank. For example, "Seattle; aerial view; scenery."
  ILGT  Lightness. Describes the changes in lightness settings on the digitizer required to produce the file. Note that the format of this information depends on hardware used.
  IMED  Medium. Describes the original subject of the file, such as, "computer image," "drawing," "lithograph," and so forth.
  INAM  Name. Stores the title of the subject of the file, such as, "Seattle From Above."
  IPLT  Palette Setting. Specifies the number of colors requested when digitizing an image, such as "256."
  IPRD  Product. Specifies the name of the title the file was originally intended for, such as "Encyclopedia of Pacific Northwest Geography."
  ISBJ  Subject. Describes the conbittents of the file, such as "Aerial view of Seattle."
  ISFT  Software. Identifies the name of the software package used to create the file, such as "Microsoft WaveEdit."
  ISHP  Sharpness. Identifies the changes in sharpness for the digitizer required to produce the file (the format depends on the hardware used).
  ISRC  Source. Identifies the name of the person or organization who supplied the original subject of the file. For example, "Trey Research."
  ISRF  Source Form. Identifies the original form of the material that was digitized, such as "slide," "paper," "map," and so forth. This is not necessarily the same as IMED.
   */
   
  uint8_t infoindex=0;
  
  do
    {
      if ((pt=memchr(haystack+infoindex+1, 0x49, span-1-infoindex)) == NULL)
        {
          
          if (globals_c.debugging) 
            {
              if (ichunk->found)
              foutput_c("[MSG]  Found %d info chunks among %d characters\n       INAM %s\n       IART %s\n       ICMT %s\n       ICOP %s\n       ICRD %s\n       IGNR %s\n", 
                          ichunk->found, span, ichunk->INAM, ichunk->IART, ichunk->ICMT, ichunk->ICOP, ichunk->ICRD, ichunk->IGNR);
              else
              foutput_c("[MSG]  Could not find info chunks among %d characters\n", span);
            
            }
          break;
        }
           
      
      infoindex=pt-haystack;
            
      if ((*(pt + 1) == 'N') && (*(pt + 2) == 'A') && (*(pt + 3) == 'M'))
          ichunk->found=read_info_chunk(pt, ichunk->INAM);
      else  
      if ((*(pt + 1) == 'A') && (*(pt + 2) == 'R') && (*(pt + 3) == 'T')) 
          ichunk->found+=read_info_chunk(pt, ichunk->IART);  
      else
      if ((*(pt + 1) == 'C'))
      {
        if ((*(pt + 2) == 'M') && (*(pt + 3) == 'T'))
            ichunk->found+=read_info_chunk(pt, ichunk->ICMT);
        else
        if ((*(pt + 2) == 'O') && (*(pt + 3) == 'P')) 
            ichunk->found+=read_info_chunk(pt, ichunk->ICOP);
        else
        if ((*(pt + 2) == 'R') && (*(pt + 3) == 'D'))
            ichunk->found+=read_info_chunk(pt, ichunk->ICRD);
      }
      else
      if ((*(pt + 1) == 'G') && (*(pt + 2) == 'N') && (*(pt + 3) == 'R')) 
            ichunk->found+=read_info_chunk(pt, ichunk->IGNR);

    }
  while (infoindex < span-4);
}
   
   
  if (globals_c.debugging)
  {
      if (span != 36) foutput_c( "[INF]  Size of header is non-standard (scanned %d characters)\n", span );
      else foutput_c("%s", "[INF]  Size of header is standard\n");

      if (span < 36)
        {
          foutput_c( "[INF]  Size of header is too short, some audio will be lost (%d bytes).\n", 36-span );
        }
  }
  
  ichunk->span=span;
  
  return;
}


FILE * secure_open(char *path, char *context)
{
  FILE *f;

  if ( (f=fopen( path, context ))  == NULL )
    {
      foutput_c("[ERR]  Could not open '%s'\n", path);
      foutput_c("%s\n", "       FILE MAY BE CORRUPTED" );
      exit(EXIT_FAILURE);
    }
  return f;

}


int end_seek(FILE *outfile)
{
  if ( fseek(outfile, 0, SEEK_END) == -1)
    {
      foutput_c( "\n%s\n", "[ERR]  Error seeking to end of output file" );
      foutput_c( "%s\n", "     File was not changed\n" );
      return(FAIL);
    }
  return(0);
}



/*********************************************************************
* Function: hexdump_header
*
* Purpose:  This function displays the first HEADER_SIZE bytes of the file
*           in both hexadecimal and ASCII
*********************************************************************/


void hexdump_header(FILE* infile, uint8_t header_size)
{
  unsigned char data[ HEX_COLUMNS ];
  size_t i, size=0, count=0, input=0;
  foutput_c( "%c", '\n' );
  do
    {

      memset(data, 0, HEX_COLUMNS);

      /* Print the base address. */
      foutput_c("%08lX:  ", (long unsigned)count);


      input= Min(header_size -count, HEX_COLUMNS);
      count+=HEX_COLUMNS;


      size = fread(data, 1, input, infile);

      if ( size ==  input)
        {

          /* Print the hex values. */
          for ( i = 0; i < HEX_COLUMNS; i++ )
            foutput_c("%02X ", data[i]);

          /* Print the characters. */
          for ( i = 0; i < HEX_COLUMNS; i++ )
            foutput_c("%c", (i < input)? (isprint(data[i]) ? data[i] : '.') : ' ');

          foutput_c("%c", '\n');
        }
      else
        {
          foutput_c("%s\n", "[ERR]  Header was not properly read by hexdump_header()");
        }

      /* break on partial buffer */
    }
  while ( count < header_size );


  foutput_c( "%c", '\n' );

}






